from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.cache_control_ephemeral_param import CacheControlEphemeralParam
    from ..models.citations_config_param import CitationsConfigParam
    from ..models.text_block_param import TextBlockParam


T = TypeVar("T", bound="SearchResultBlockParam")


@_attrs_define
class SearchResultBlockParam:
    """
    Attributes:
        content (list['TextBlockParam']):
        source (str):
        title (str):
        type_ (Literal['search_result']):
        cache_control (Union['CacheControlEphemeralParam', None, Unset]):
        citations (Union[Unset, CitationsConfigParam]):
    """

    content: list["TextBlockParam"]
    source: str
    title: str
    type_: Literal["search_result"]
    cache_control: Union["CacheControlEphemeralParam", None, Unset] = UNSET
    citations: Union[Unset, "CitationsConfigParam"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.cache_control_ephemeral_param import CacheControlEphemeralParam

        content = []
        for content_item_data in self.content:
            content_item = content_item_data.to_dict()
            content.append(content_item)

        source = self.source

        title = self.title

        type_ = self.type_

        cache_control: Union[None, Unset, dict[str, Any]]
        if isinstance(self.cache_control, Unset):
            cache_control = UNSET
        elif isinstance(self.cache_control, CacheControlEphemeralParam):
            cache_control = self.cache_control.to_dict()
        else:
            cache_control = self.cache_control

        citations: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.citations, Unset):
            citations = self.citations.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "content": content,
                "source": source,
                "title": title,
                "type": type_,
            }
        )
        if cache_control is not UNSET:
            field_dict["cache_control"] = cache_control
        if citations is not UNSET:
            field_dict["citations"] = citations

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.cache_control_ephemeral_param import CacheControlEphemeralParam
        from ..models.citations_config_param import CitationsConfigParam
        from ..models.text_block_param import TextBlockParam

        d = dict(src_dict)
        content = []
        _content = d.pop("content")
        for content_item_data in _content:
            content_item = TextBlockParam.from_dict(content_item_data)

            content.append(content_item)

        source = d.pop("source")

        title = d.pop("title")

        type_ = cast(Literal["search_result"], d.pop("type"))
        if type_ != "search_result":
            raise ValueError(f"type must match const 'search_result', got '{type_}'")

        def _parse_cache_control(data: object) -> Union["CacheControlEphemeralParam", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                cache_control_type_0 = CacheControlEphemeralParam.from_dict(data)

                return cache_control_type_0
            except:  # noqa: E722
                pass
            return cast(Union["CacheControlEphemeralParam", None, Unset], data)

        cache_control = _parse_cache_control(d.pop("cache_control", UNSET))

        _citations = d.pop("citations", UNSET)
        citations: Union[Unset, CitationsConfigParam]
        if isinstance(_citations, Unset):
            citations = UNSET
        else:
            citations = CitationsConfigParam.from_dict(_citations)

        search_result_block_param = cls(
            content=content,
            source=source,
            title=title,
            type_=type_,
            cache_control=cache_control,
            citations=citations,
        )

        search_result_block_param.additional_properties = d
        return search_result_block_param

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .base_64_image_source_param import Base64ImageSourceParam
from .base_64_image_source_param_media_type import Base64ImageSourceParamMediaType
from .cache_control_ephemeral_param import CacheControlEphemeralParam
from .canvas_commit_message_event import CanvasCommitMessageEvent
from .canvas_data import CanvasData
from .canvas_data_nodes import CanvasDataNodes
from .canvas_list_response import CanvasListResponse
from .canvas_summary import CanvasSummary
from .canvas_summary_meta import CanvasSummaryMeta
from .canvas_update_message_event import CanvasUpdateMessageEvent
from .citation_char_location_param import CitationCharLocationParam
from .citation_content_block_location_param import CitationContentBlockLocationParam
from .citation_page_location_param import CitationPageLocationParam
from .citation_search_result_location_param import CitationSearchResultLocationParam
from .citation_web_search_result_location_param import CitationWebSearchResultLocationParam
from .citations_config_param import CitationsConfigParam
from .commit_message_request import CommitMessageRequest
from .create_canvas_request import CreateCanvasRequest
from .create_canvas_response import CreateCanvasResponse
from .create_message_response import CreateMessageResponse
from .delete_canvas_response import DeleteCanvasResponse
from .dimension import Dimension
from .get_canvas_response import GetCanvasResponse
from .health_check_response import HealthCheckResponse
from .health_check_response_server_type import HealthCheckResponseServerType
from .http_validation_error import HTTPValidationError
from .image_block_param import ImageBlockParam
from .message import Message
from .message_node import MessageNode
from .message_node_meta import MessageNodeMeta
from .message_role import MessageRole
from .position import Position
from .search_result_block_param import SearchResultBlockParam
from .sse_canvas_created_event import SSECanvasCreatedEvent
from .sse_canvas_deleted_event import SSECanvasDeletedEvent
from .sse_canvas_deleted_event_data import SSECanvasDeletedEventData
from .sse_canvas_updated_event import SSECanvasUpdatedEvent
from .sse_documentation_response import SSEDocumentationResponse
from .sse_error_event import SSEErrorEvent
from .sse_error_event_data import SSEErrorEventData
from .sse_heartbeat_event import SSEHeartbeatEvent
from .sse_message_committed_event import SSEMessageCommittedEvent
from .sse_message_deleted_event import SSEMessageDeletedEvent
from .sse_message_deleted_event_data import SSEMessageDeletedEventData
from .sse_message_updated_event import SSEMessageUpdatedEvent
from .text_block_param import TextBlockParam
from .tool_result_block_param import ToolResultBlockParam
from .tool_use_block_param import ToolUseBlockParam
from .update_message_request import UpdateMessageRequest
from .url_image_source_param import URLImageSourceParam
from .validation_error import ValidationError

__all__ = (
    "Base64ImageSourceParam",
    "Base64ImageSourceParamMediaType",
    "CacheControlEphemeralParam",
    "CanvasCommitMessageEvent",
    "CanvasData",
    "CanvasDataNodes",
    "CanvasListResponse",
    "CanvasSummary",
    "CanvasSummaryMeta",
    "CanvasUpdateMessageEvent",
    "CitationCharLocationParam",
    "CitationContentBlockLocationParam",
    "CitationPageLocationParam",
    "CitationsConfigParam",
    "CitationSearchResultLocationParam",
    "CitationWebSearchResultLocationParam",
    "CommitMessageRequest",
    "CreateCanvasRequest",
    "CreateCanvasResponse",
    "CreateMessageResponse",
    "DeleteCanvasResponse",
    "Dimension",
    "GetCanvasResponse",
    "HealthCheckResponse",
    "HealthCheckResponseServerType",
    "HTTPValidationError",
    "ImageBlockParam",
    "Message",
    "MessageNode",
    "MessageNodeMeta",
    "MessageRole",
    "Position",
    "SearchResultBlockParam",
    "SSECanvasCreatedEvent",
    "SSECanvasDeletedEvent",
    "SSECanvasDeletedEventData",
    "SSECanvasUpdatedEvent",
    "SSEDocumentationResponse",
    "SSEErrorEvent",
    "SSEErrorEventData",
    "SSEHeartbeatEvent",
    "SSEMessageCommittedEvent",
    "SSEMessageDeletedEvent",
    "SSEMessageDeletedEventData",
    "SSEMessageUpdatedEvent",
    "TextBlockParam",
    "ToolResultBlockParam",
    "ToolUseBlockParam",
    "UpdateMessageRequest",
    "URLImageSourceParam",
    "ValidationError",
)



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="CitationPageLocationParam")


@_attrs_define
class CitationPageLocationParam:
    """
    Attributes:
        cited_text (str):
        document_index (int):
        document_title (Union[None, str]):
        end_page_number (int):
        start_page_number (int):
        type_ (Literal['page_location']):
    """

    cited_text: str
    document_index: int
    document_title: Union[None, str]
    end_page_number: int
    start_page_number: int
    type_: Literal["page_location"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cited_text = self.cited_text

        document_index = self.document_index

        document_title: Union[None, str]
        document_title = self.document_title

        end_page_number = self.end_page_number

        start_page_number = self.start_page_number

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "cited_text": cited_text,
                "document_index": document_index,
                "document_title": document_title,
                "end_page_number": end_page_number,
                "start_page_number": start_page_number,
                "type": type_,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        cited_text = d.pop("cited_text")

        document_index = d.pop("document_index")

        def _parse_document_title(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        document_title = _parse_document_title(d.pop("document_title"))

        end_page_number = d.pop("end_page_number")

        start_page_number = d.pop("start_page_number")

        type_ = cast(Literal["page_location"], d.pop("type"))
        if type_ != "page_location":
            raise ValueError(f"type must match const 'page_location', got '{type_}'")

        citation_page_location_param = cls(
            cited_text=cited_text,
            document_index=document_index,
            document_title=document_title,
            end_page_number=end_page_number,
            start_page_number=start_page_number,
            type_=type_,
        )

        citation_page_location_param.additional_properties = d
        return citation_page_location_param

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Type definitions for llm_canvas.

This module contains all TypedDict definitions used throughout the llm_canvas package,
including API request/response types and core data structures.
"""

from __future__ import annotations

from typing import Any, Literal, TypedDict, Union

try:
    from typing import NotRequired
except ImportError:
    from typing_extensions import NotRequired

from anthropic.types import ImageBlockParam, TextBlockParam, ToolResultBlockParam, ToolUseBlockParam

# ---- Core Data Types ----

# Union type for message blocks matching TypeScript
MessageBlock = Union[TextBlockParam, ToolUseBlockParam, ToolResultBlockParam, ImageBlockParam]


class Message(TypedDict):
    """Message structure for canvas conversations."""

    content: Union[str, list[MessageBlock]]
    role: Literal["user", "assistant", "system"]


class Position(TypedDict):
    """Position information for a message node."""

    x: float
    y: float


class Dimension(TypedDict):
    """Dimension information for a message node."""

    width: float
    height: float


class MessageNode(TypedDict):
    """Node in the canvas conversation graph."""

    id: str
    message: Message
    child_ids: list[str]
    parent_id: NotRequired[str]
    meta: NotRequired[dict[str, Any]]
    position: NotRequired[Position]
    dimension: NotRequired[Dimension]


class CanvasSummary(TypedDict):
    """Summary information about a canvas."""

    canvas_id: str
    created_at: float
    root_ids: list[str]
    node_count: int
    title: Union[str, None]
    description: Union[str, None]
    meta: dict[str, Any]


class CanvasData(TypedDict):
    """Complete canvas data structure."""

    title: Union[str, None]
    last_updated: Union[float, None]
    description: Union[str, None]
    canvas_id: str
    created_at: float
    nodes: dict[str, MessageNode]


CanvasEventType = Literal["commit_message", "update_message", "delete_message"]


class CanvasCommitMessageEvent(TypedDict):
    """Event data for canvas message commits."""

    event_type: Literal["commit_message"]
    canvas_id: str
    timestamp: float
    data: MessageNode


class CanvasUpdateMessageEvent(TypedDict):
    """Event data for canvas message updates."""

    event_type: Literal["update_message"]
    canvas_id: str
    timestamp: float
    data: MessageNode


class CanvasDeleteMessageEvent(TypedDict):
    """Event data for canvas message deletions."""

    event_type: Literal["delete_message"]
    canvas_id: str
    timestamp: float
    data: str  # Node ID that was deleted


CanvasEvent = Union[CanvasCommitMessageEvent, CanvasUpdateMessageEvent, CanvasDeleteMessageEvent]


class BranchInfo(TypedDict):
    """Information about a canvas branch."""

    name: str
    description: Union[str, None]
    head_node_id: Union[str, None]
    created_at: float



"""Canvas Client - A high-level interface for managing canvases and running the server.

This module provides a simplified, one-stop solution for users to:
- Create and manage multiple canvases
- Add canvases to a registry
- Run the web UI/API server
- Handle canvas operations in a unified way
"""

# ruff: noqa: BLE001
# ruff: noqa: TRY301

from __future__ import annotations

import logging
import threading
from typing import Union

from httpx import Timeout

from llm_canvas.canvas_registry import CanvasRegistry
from llm_canvas.types import CanvasCommitMessageEvent, CanvasEvent, CanvasUpdateMessageEvent
from llm_canvas_generated_client.llm_canvas_api_client import Client
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    commit_message_api_v1_canvas_canvas_id_messages_post as commit_message_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    create_canvas_api_v1_canvas_post as create_canvas_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    delete_canvas_api_v1_canvas_canvas_id_delete as delete_canvas_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    get_canvas_api_v1_canvas_get as get_canvas_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    health_check_api_v1_health_get as health_check_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    list_canvases_api_v1_canvas_list_get as list_canvases_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.api.v1 import (
    update_message_api_v1_canvas_canvas_id_messages_message_id_put as update_message_api,
)
from llm_canvas_generated_client.llm_canvas_api_client.models.canvas_commit_message_event import (
    CanvasCommitMessageEvent as GeneratedCanvasCommitMessageEvent,
)
from llm_canvas_generated_client.llm_canvas_api_client.models.canvas_update_message_event import (
    CanvasUpdateMessageEvent as GeneratedCanvasUpdateMessageEvent,
)
from llm_canvas_generated_client.llm_canvas_api_client.models.commit_message_request import (
    CommitMessageRequest,
)
from llm_canvas_generated_client.llm_canvas_api_client.models.create_canvas_request import CreateCanvasRequest
from llm_canvas_generated_client.llm_canvas_api_client.models.create_canvas_response import CreateCanvasResponse
from llm_canvas_generated_client.llm_canvas_api_client.models.http_validation_error import HTTPValidationError
from llm_canvas_generated_client.llm_canvas_api_client.models.update_message_request import UpdateMessageRequest

from .canvas import Canvas, CanvasData, CanvasSummary

logger = logging.getLogger(__name__)


class CanvasClient:
    """High-level client for managing canvases and running the server.

    This class provides a unified interface for:
    - Creating and managing multiple canvases
    - Running the web UI/API server
    - Handling canvas operations

    Example:
        client = CanvasClient()
        canvas = client.create_canvas("My Chat", "A conversation about AI")
        client.add_message(canvas.canvas_id, "Hello!", "user")
    """

    def __init__(self, server_host: str = "127.0.0.1", server_port: int = 8000) -> None:
        self.registry = CanvasRegistry()
        self._server_thread: Union[threading.Thread, None] = None
        self._server_running = False
        self.server_host = server_host
        self.server_port = server_port
        # Initialize the API client
        base_url = f"http://{server_host}:{server_port}"
        self._api_client = Client(base_url=base_url, timeout=Timeout(10.0))

        # Event tracking for canvases
        self._event_lock = threading.Lock()

        if not self.check_server_health():
            self._prompt_user_to_start_server()

    def _on_canvas_event(self, event: CanvasEvent) -> None:
        """Internal event handler that forwards canvas events to registered listeners and calls API endpoints."""
        # Call API endpoints for commit and update events if server is available
        if self._ensure_server_running():
            try:
                if event["event_type"] == "commit_message":
                    self._call_commit_message_api(event)
                elif event["event_type"] == "update_message":
                    self._call_update_message_api(event)
                # Ignore delete_message events for now
            except Exception:
                logger.exception("Error calling API endpoint for canvas event")

    def _call_commit_message_api(self, event: CanvasCommitMessageEvent) -> None:
        """Call the commit message API endpoint."""
        canvas_id = event["canvas_id"]

        try:
            request = CommitMessageRequest(data=GeneratedCanvasCommitMessageEvent.from_dict(event))
            response = commit_message_api.sync(canvas_id=canvas_id, client=self._api_client, body=request)

            if response:
                logger.debug("Successfully called commit message API for canvas %s", canvas_id)
            else:
                logger.warning("Failed to call commit message API")

        except Exception as e:
            logger.warning("Failed to call commit message API: %s", e)

    def _call_update_message_api(self, event: CanvasUpdateMessageEvent) -> None:
        """Call the update message API endpoint."""
        canvas_id = event["canvas_id"]
        message_id = event["data"]["id"]

        try:
            request = UpdateMessageRequest(GeneratedCanvasUpdateMessageEvent.from_dict(event))
            response = update_message_api.sync(
                canvas_id=canvas_id, message_id=message_id, client=self._api_client, body=request
            )

            if response:
                logger.debug("Successfully called update message API for canvas %s", canvas_id)
            else:
                logger.warning("Failed to call update message API")

        except Exception as e:
            logger.warning("Failed to call update message API: %s", e)

    def _setup_canvas_event_tracking(self, canvas: Canvas) -> None:
        """Set up event tracking for a canvas by adding our event listener."""
        canvas.add_event_listener(self._on_canvas_event)

    def check_server_health(self) -> bool:
        """Check if the server is running and healthy.

        Returns:
            True if server is running and healthy, False otherwise
        """
        try:
            response = health_check_api.sync(client=self._api_client)
        except Exception:
            # Any exception means server is not reachable
            return False
        else:
            return response is not None and response.status == "healthy"

    def _prompt_user_to_start_server(self) -> None:
        """Prompt user to start the server manually."""
        print("❌ Canvas server is not running!")
        print("\n📋 To start the local canvas server, run one of these commands:")
        print(f"   llm-canvas server --host {self.server_host} --port {self.server_port}")
        print("   llm-canvas server  # (uses default host and port)")
        print("\n📖 For more information, see: doc/start_canvas_server.md")
        print(f"\n🌐 Once started, the server will be available at: http://{self.server_host}:{self.server_port}")

    def _ensure_server_running(self) -> bool:
        """Ensure server is running, prompt user if not.

        Returns:
            True if server is running, False if user needs to start it manually
        """
        if self.check_server_health():
            return True

        self._prompt_user_to_start_server()
        return False

    def create_canvas(
        self,
        title: Union[str, None] = None,
        description: Union[str, None] = None,
    ) -> Canvas:
        """Create a new canvas and add it to the registry.

        Args:
            title: Optional title for the canvas
            description: Optional description for the canvas

        Returns:
            The created Canvas instance

        Raises:
            RuntimeError: If server is not running and user needs to start it manually
        """
        if not self._ensure_server_running():
            error_msg = "Canvas server is not running. Please start the server manually using 'llm-canvas server'."
            raise RuntimeError(error_msg)

        # Call API to create canvas
        try:
            request = CreateCanvasRequest(title=title, description=description)
            response = create_canvas_api.sync(client=self._api_client, body=request)

            if isinstance(response, CreateCanvasResponse):
                created_canvas_id = response.canvas_id

                # Fetch the full canvas data
                canvas = self.get_canvas(created_canvas_id)
                if canvas:
                    logger.info("Created canvas via API: %s - %s", created_canvas_id, title)
                    return canvas

                msg = f"Failed to retrieve created canvas {created_canvas_id}"
                raise RuntimeError(msg)

            msg = "Failed to create canvas: No response from API"
            raise RuntimeError(msg)

        except Exception as e:
            msg = f"Failed to create canvas via API: {e}"
            raise RuntimeError(msg) from e

    def get_canvas(self, canvas_id: str) -> Union[Canvas, None]:
        """Get a canvas by ID.

        Args:
            canvas_id: The canvas ID to retrieve

        Returns:
            The Canvas instance if found, None otherwise
        """
        # Call API to get canvas
        try:
            canvas_data_response = get_canvas_api.sync(client=self._api_client, canvas_id=canvas_id)

            if not isinstance(canvas_data_response, HTTPValidationError) and canvas_data_response is not None:
                # Convert API response to Canvas object
                canvas = Canvas.from_canvas_data(
                    CanvasData(
                        title=canvas_data_response.data.title,
                        description=canvas_data_response.data.description,
                        created_at=canvas_data_response.data.created_at,
                        last_updated=canvas_data_response.data.last_updated,
                        nodes=canvas_data_response.data.nodes.to_dict(),
                        canvas_id=canvas_data_response.data.canvas_id,
                    )
                )
                self._setup_canvas_event_tracking(canvas)
                return canvas
            return None

        except Exception as e:
            logger.warning("Failed to get canvas %s via API: %s", canvas_id, e)
            return None

    def list_canvases(self) -> list[Canvas]:
        """List all canvases in the registry.

        Returns:
            List of all Canvas instances
        """
        if not self._ensure_server_running():
            return self.registry.list()

        # Call API to get canvas list and then fetch each canvas
        try:
            response = list_canvases_api.sync(client=self._api_client)

            if response:
                canvases = []
                for summary in response.canvases:
                    canvas = self.get_canvas(summary.canvas_id)
                    if canvas:
                        canvases.append(canvas)
                return canvases
            logger.warning("Failed to list canvases: No response from API")
            return []

        except Exception as e:
            logger.warning("Failed to list canvases via API: %s", e)
            return []

    def get_canvas_summaries(self) -> list[CanvasSummary]:
        """Get summaries of all canvases.

        Returns:
            List of CanvasSummary objects
        """
        if not self._ensure_server_running():
            summaries = []
            for canvas in self.registry.list():
                summary = canvas.to_summary()
                # Update with registry's last_updated time
                if registry_updated := self.registry.last_updated(canvas.canvas_id):
                    summary["meta"]["last_updated"] = registry_updated
                summaries.append(summary)
            return summaries

        # Call API to get canvas summaries
        try:
            response = list_canvases_api.sync(client=self._api_client)

            if response:
                return [
                    {
                        "canvas_id": c.canvas_id,
                        "created_at": c.created_at,
                        "root_ids": c.root_ids,
                        "node_count": c.node_count,
                        "title": c.title,
                        "description": c.description,
                        "meta": c.meta.to_dict() if c.meta else {},
                    }
                    for c in response.canvases
                ]
            logger.warning("Failed to get canvas summaries: No response from API")
            return []

        except Exception as e:
            logger.warning("Failed to get canvas summaries via API: %s", e)
            return []

    def get_canvas_data(self, canvas_id: str) -> Union[CanvasData, None]:
        """Get canvas data in the standard format.

        Args:
            canvas_id: The canvas ID to retrieve

        Returns:
            CanvasData if found, None otherwise
        """
        if not self._ensure_server_running():
            canvas = self.registry.get(canvas_id)
            if not canvas:
                return None
            return canvas.to_canvas_data()

        # Call API to get canvas data
        try:
            canvas_data_response = get_canvas_api.sync(client=self._api_client, canvas_id=canvas_id)
            if not isinstance(canvas_data_response, HTTPValidationError) and canvas_data_response is not None:
                return CanvasData(
                    canvas_id=canvas_data_response.data.canvas_id,
                    title=canvas_data_response.data.title,
                    description=canvas_data_response.data.description,
                    created_at=canvas_data_response.data.created_at,
                    last_updated=canvas_data_response.data.last_updated,
                    nodes=canvas_data_response.data.nodes.to_dict(),
                )
            return None

        except Exception as e:
            logger.warning("Failed to get canvas data %s via API: %s", canvas_id, e)
            return None

    def remove_canvas(self, canvas_id: str) -> bool:
        """Remove a canvas from the registry.

        Args:
            canvas_id: The canvas ID to remove

        Returns:
            True if removed successfully, False if not found
        """
        if not self._ensure_server_running():
            removed = self.registry.remove(canvas_id)
            if removed:
                logger.info("Removed canvas: %s", canvas_id)
            return removed

        # Call API to delete canvas
        try:
            response = delete_canvas_api.sync(canvas_id=canvas_id, client=self._api_client)

            if response:
                logger.info("Removed canvas via API: %s", canvas_id)
                return True
            return False

        except Exception as e:
            logger.warning("Failed to remove canvas %s via API: %s", canvas_id, e)
            return False

    def __len__(self) -> int:
        """Return the number of canvases in the registry."""
        return len(self.registry.list())

    def __repr__(self) -> str:
        """Return a string representation of the client."""
        return f"CanvasClient(canvases={len(self)})"



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.dimension import Dimension
    from ..models.message import Message
    from ..models.message_node_meta import MessageNodeMeta
    from ..models.position import Position


T = TypeVar("T", bound="MessageNode")


@_attrs_define
class MessageNode:
    """Node in the canvas conversation graph.

    Attributes:
        id (str):
        message (Message): Message structure for canvas conversations.
        child_ids (list[str]):
        parent_id (Union[Unset, str]):
        meta (Union[Unset, MessageNodeMeta]):
        position (Union[Unset, Position]): Position information for a message node.
        dimension (Union[Unset, Dimension]): Dimension information for a message node.
    """

    id: str
    message: "Message"
    child_ids: list[str]
    parent_id: Union[Unset, str] = UNSET
    meta: Union[Unset, "MessageNodeMeta"] = UNSET
    position: Union[Unset, "Position"] = UNSET
    dimension: Union[Unset, "Dimension"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        message = self.message.to_dict()

        child_ids = self.child_ids

        parent_id = self.parent_id

        meta: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.meta, Unset):
            meta = self.meta.to_dict()

        position: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.position, Unset):
            position = self.position.to_dict()

        dimension: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.dimension, Unset):
            dimension = self.dimension.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "message": message,
                "child_ids": child_ids,
            }
        )
        if parent_id is not UNSET:
            field_dict["parent_id"] = parent_id
        if meta is not UNSET:
            field_dict["meta"] = meta
        if position is not UNSET:
            field_dict["position"] = position
        if dimension is not UNSET:
            field_dict["dimension"] = dimension

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.dimension import Dimension
        from ..models.message import Message
        from ..models.message_node_meta import MessageNodeMeta
        from ..models.position import Position

        d = dict(src_dict)
        id = d.pop("id")

        message = Message.from_dict(d.pop("message"))

        child_ids = cast(list[str], d.pop("child_ids"))

        parent_id = d.pop("parent_id", UNSET)

        _meta = d.pop("meta", UNSET)
        meta: Union[Unset, MessageNodeMeta]
        if isinstance(_meta, Unset):
            meta = UNSET
        else:
            meta = MessageNodeMeta.from_dict(_meta)

        _position = d.pop("position", UNSET)
        position: Union[Unset, Position]
        if isinstance(_position, Unset):
            position = UNSET
        else:
            position = Position.from_dict(_position)

        _dimension = d.pop("dimension", UNSET)
        dimension: Union[Unset, Dimension]
        if isinstance(_dimension, Unset):
            dimension = UNSET
        else:
            dimension = Dimension.from_dict(_dimension)

        message_node = cls(
            id=id,
            message=message,
            child_ids=child_ids,
            parent_id=parent_id,
            meta=meta,
            position=position,
            dimension=dimension,
        )

        message_node.additional_properties = d
        return message_node

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""CLI entry point for llm_canvas.

Provides command-line interface for starting the local server.
"""

from __future__ import annotations

import argparse
import sys

from ._server import start_local_server


def main() -> None:  # pragma: no cover - CLI utility
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(
        description="LLM Canvas - Visualize LLM conversations as a navigable canvas",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  llm-canvas server                    # Start local server on default port 8000
  llm-canvas server --port 3000       # Start local server on port 3000
  llm-canvas server --log-level debug # Start with debug logging
        """,
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Server subcommand
    server_parser = subparsers.add_parser("server", help="Start the local LLM Canvas server")
    server_parser.add_argument("--host", default="127.0.0.1", help="Host to serve on (default: 127.0.0.1)")
    server_parser.add_argument("--port", type=int, default=8000, help="Port to serve on (default: 8000)")
    server_parser.add_argument(
        "--log-level", default="info", choices=["debug", "info", "warning", "error"], help="Set logging level (default: info)"
    )

    args = parser.parse_args()

    if args.command == "server":
        start_local_server(host=args.host, port=args.port, log_level=args.log_level)
    elif args.command is None:
        # Default to server if no subcommand provided
        print("No command specified. Starting local server...")
        print("Use 'llm-canvas server --help' for more options.")
        start_local_server()
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":  # pragma: no cover
    main()



"""Event system for Server-Side Events (SSE) in llm_canvas.

This module provides event dispatching and connection management for SSE endpoints.
"""

from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import time
from collections import defaultdict
from collections.abc import AsyncGenerator

from llm_canvas.types import (
    CanvasSummary,
    MessageNode,
)

from ._types import (
    SSECanvasCreatedEvent,
    SSECanvasDeletedEvent,
    SSECanvasEvent,
    SSECanvasUpdatedEvent,
    SSEGlobalEvent,
    SSEMessageCommittedEvent,
    SSEMessageDeletedEvent,
    SSEMessageUpdatedEvent,
)

logger = logging.getLogger(__name__)


class SSEEventDispatcher:
    """Manages SSE connections and event distribution."""

    def __init__(self) -> None:
        # Global canvas events (for /canvas/sse)
        self._global_connections: set[asyncio.Queue[str]] = set()

        # Per-canvas events (for /canvas/{canvas_id}/sse)
        self._canvas_connections: dict[str, set[asyncio.Queue[str]]] = defaultdict(set)

        # Lock for thread-safe operations
        self._lock = asyncio.Lock()

        # Shutdown event to signal all connections to close
        self._shutdown_event = asyncio.Event()

    async def add_global_connection(self, queue: asyncio.Queue[str]) -> None:
        """Add a connection for global canvas events."""
        async with self._lock:
            self._global_connections.add(queue)
            logger.info(f"Added global SSE connection. Total: {len(self._global_connections)}")

    async def remove_global_connection(self, queue: asyncio.Queue[str]) -> None:
        """Remove a connection for global canvas events."""
        async with self._lock:
            self._global_connections.discard(queue)
            logger.info(f"Removed global SSE connection. Total: {len(self._global_connections)}")

    async def add_canvas_connection(self, canvas_id: str, queue: asyncio.Queue[str]) -> None:
        """Add a connection for specific canvas events."""
        async with self._lock:
            self._canvas_connections[canvas_id].add(queue)
            logger.info(f"Added canvas {canvas_id} SSE connection. Total: {len(self._canvas_connections[canvas_id])}")

    async def remove_canvas_connection(self, canvas_id: str, queue: asyncio.Queue[str]) -> None:
        """Remove a connection for specific canvas events."""
        async with self._lock:
            self._canvas_connections[canvas_id].discard(queue)
            if not self._canvas_connections[canvas_id]:
                del self._canvas_connections[canvas_id]
            logger.info(f"Removed canvas {canvas_id} SSE connection")

    async def broadcast_global_event(self, event_data: SSEGlobalEvent) -> None:
        """Broadcast an event to all global connections."""
        if not self._global_connections:
            return

        message = f"event: {event_data['type']}\ndata: {json.dumps(event_data)}\n\n"

        async with self._lock:
            disconnected_queues = set()

            # Process each queue and collect results
            for queue in self._global_connections:
                if queue.full():
                    logger.warning("SSE queue full, marking for removal")
                    disconnected_queues.add(queue)
                    continue

                try:
                    queue.put_nowait(message)
                except (RuntimeError, ValueError):
                    logger.exception("Error sending to SSE queue")
                    disconnected_queues.add(queue)

            # Remove disconnected queues
            for queue in disconnected_queues:
                self._global_connections.discard(queue)

    async def broadcast_canvas_event(self, canvas_id: str, event_data: SSECanvasEvent) -> None:
        """Broadcast an event to all connections for a specific canvas."""
        if canvas_id not in self._canvas_connections:
            return

        message = f"event: {event_data['type']}\ndata: {json.dumps(event_data)}\n\n"

        async with self._lock:
            disconnected_queues = set()

            # Process each queue and collect results
            for queue in self._canvas_connections[canvas_id]:
                if queue.full():
                    logger.warning("SSE queue full, marking for removal")
                    disconnected_queues.add(queue)
                    continue

                try:
                    queue.put_nowait(message)
                except (RuntimeError, ValueError):
                    logger.exception("Error sending to SSE queue")
                    disconnected_queues.add(queue)

            # Remove disconnected queues
            for queue in disconnected_queues:
                self._canvas_connections[canvas_id].discard(queue)

    async def canvas_created(self, canvas_summary: CanvasSummary) -> None:
        """Broadcast that a canvas was created."""
        await self.broadcast_global_event(
            SSECanvasCreatedEvent(type="canvas_created", timestamp=time.time(), data=canvas_summary)
        )

    async def canvas_updated(self, canvas_summary: CanvasSummary) -> None:
        """Broadcast that a canvas was updated."""
        await self.broadcast_global_event(
            SSECanvasUpdatedEvent(type="canvas_updated", timestamp=time.time(), data=canvas_summary)
        )

    async def canvas_deleted(self, canvas_id: str) -> None:
        """Broadcast that a canvas was deleted."""
        await self.broadcast_global_event(
            SSECanvasDeletedEvent(type="canvas_deleted", timestamp=time.time(), data={"canvas_id": canvas_id})
        )

    async def message_committed(self, canvas_id: str, message_data: MessageNode) -> None:
        """Broadcast that a message was committed to a canvas."""
        await self.broadcast_canvas_event(
            canvas_id,
            SSEMessageCommittedEvent(type="message_committed", timestamp=time.time(), canvas_id=canvas_id, data=message_data),
        )

    async def message_updated(self, canvas_id: str, message_data: MessageNode) -> None:
        """Broadcast that a message was updated in a canvas."""
        await self.broadcast_canvas_event(
            canvas_id,
            SSEMessageUpdatedEvent(type="message_updated", timestamp=time.time(), canvas_id=canvas_id, data=message_data),
        )

    async def message_deleted(self, canvas_id: str, message_id: str) -> None:
        """Broadcast that a message was deleted from a canvas."""
        await self.broadcast_canvas_event(
            canvas_id,
            SSEMessageDeletedEvent(
                type="message_deleted", timestamp=time.time(), canvas_id=canvas_id, data={"message_id": message_id}
            ),
        )

    async def shutdown(self) -> None:
        """Signal all connections to close and clean up."""
        logger.info("Shutting down SSE event dispatcher")
        self._shutdown_event.set()

        async with self._lock:
            # Send shutdown message to all connections
            shutdown_message = 'event: shutdown\ndata: {"message": "Server shutting down"}\n\n'

            # Notify all global connections
            for queue in self._global_connections.copy():
                with contextlib.suppress(RuntimeError, ValueError):
                    queue.put_nowait(shutdown_message)

            # Notify all canvas connections
            for canvas_connections in self._canvas_connections.values():
                for queue in canvas_connections.copy():
                    with contextlib.suppress(RuntimeError, ValueError):
                        queue.put_nowait(shutdown_message)

            # Clear all connections
            self._global_connections.clear()
            self._canvas_connections.clear()

        logger.info("SSE event dispatcher shutdown complete")

    @property
    def is_shutdown(self) -> bool:
        """Check if shutdown has been initiated."""
        return self._shutdown_event.is_set()

    async def wait_for_shutdown(self) -> None:
        """Wait for shutdown event to be set."""
        await self._shutdown_event.wait()


# Global event dispatcher instance
_event_dispatcher: SSEEventDispatcher | None = None


def get_event_dispatcher() -> SSEEventDispatcher:
    """Get the global event dispatcher instance."""
    global _event_dispatcher  # noqa: PLW0603
    if _event_dispatcher is None:
        _event_dispatcher = SSEEventDispatcher()
    return _event_dispatcher


async def create_sse_stream(queue: asyncio.Queue[str]) -> AsyncGenerator[str, None]:
    """Create an SSE stream from a queue."""
    event_dispatcher = get_event_dispatcher()

    try:
        while not event_dispatcher.is_shutdown:
            # Create tasks for message, timeout, and shutdown
            message_task = asyncio.create_task(queue.get())
            timeout_task = asyncio.create_task(asyncio.sleep(10.0))
            shutdown_task = asyncio.create_task(event_dispatcher.wait_for_shutdown())

            done, pending = await asyncio.wait([message_task, timeout_task, shutdown_task], return_when=asyncio.FIRST_COMPLETED)

            # Cancel any pending tasks
            for task in pending:
                task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await task

            if shutdown_task in done:
                # Shutdown requested, break the loop
                logger.info("SSE stream received shutdown signal")
                break
            elif message_task in done:
                # We received a message
                message = message_task.result()
                # Check if it's a shutdown message
                if "event: shutdown" in message:
                    yield message
                    break
                yield message
            else:
                # Timeout occurred, send heartbeat
                yield "event: heartbeat\ndata: {}\n\n"

    except asyncio.CancelledError:
        logger.info("SSE stream cancelled")
        raise
    except (RuntimeError, ValueError):
        logger.exception("Error in SSE stream")
        yield f"event: error\ndata: {json.dumps({'error': 'stream error'})}\n\n"
    finally:
        logger.info("SSE stream ended")
